/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.

     Portions of this code were adopted from the
     gnome-system-monitor v2.0.5, (C) 2001 Kevin Vandersloot


     Todo:
     - make StatEntry more expressive; we want to have
       "delta" noise relative to "delta" total decrypted,
       it should be possible to capture something like that
       in the struct.
     - add any more StatEntries, update menu accordingly.
*/

#include "gnunet_afs_esed2.h"
#include "statistics.h"
#include "helper.h"
#include "main.h"
#include <glib.h>

typedef struct {
  char * statName;
  char * paneName;
  char * frameName;
  long long normalizeBy;  
} StatEntry;

static StatEntry stats[] = {
  { "# currently connected nodes",
    "Connectivity",
    "# connected nodes (100% = 100 nodes)",
    100,
  }, 
  { "% of allowed cpu load",
    "CPU load",
    "% of allowed load",
    100,
  },
};

static void statClose(void);

static GtkItemFactoryEntry statWindowMenu[] = {
  { "/Close display",   NULL,   statClose,       0, "<Item>" }
};
static gint statWindowMenuItems 
  = sizeof (statWindowMenu) / sizeof (statWindowMenu[0]);

typedef struct {
  gint type;
  guint speed;
  guint draw_width, draw_height;
  guint num_points; 
  guint allocated;  
  GdkColor *colors;
  gfloat **data, **odata;
  guint data_size;
  guint *pos; 
  gint colors_allocated;
  GtkWidget *main_widget;
  GtkWidget *disp;
  GtkWidget *label;
  GdkPixmap *pixmap;
  GdkGC *gc;
  int timer_index;  
  gboolean draw;
  GNUNET_TCP_SOCKET * sock;
  int statIdx;
} LoadGraph ;

typedef struct {
  gint            graph_update_interval;
  GdkColor        mem_color;
  GdkColor        bg_color;
  GdkColor        frame_color;
} ProcConfig;

typedef struct ProcData {
  ProcConfig      config;
  LoadGraph       *mem_graph;
  int statIdx;
} ProcData;

#define GNOME_PAD_SMALL 2
#define FRAME_WIDTH 0

/**
 * Remove the active page from the notebook.
 **/
static void statClose(void) {
  gint pagenr;

  pagenr = gtk_notebook_get_current_page(notebook);
  gtk_notebook_remove_page(notebook, pagenr);
  /* Need to refresh the widget --
     This forces the widget to redraw itself. */
  gtk_widget_draw(GTK_WIDGET(notebook), NULL);
}

/* Redraws the backing pixmap for the load graph and updates the window */
static void load_graph_draw (LoadGraph *g) {
  guint i;
  gint dely;
  float delx;
  
  if (!g->disp->window)
    return;
  
  g->draw_width = g->disp->allocation.width;
  g->draw_height = g->disp->allocation.height;
  
  if (!g->pixmap)
    g->pixmap = gdk_pixmap_new (g->disp->window,
				g->draw_width, g->draw_height,
				gtk_widget_get_visual (g->disp)->depth);
  
  /* Create GC if necessary. */
  if (!g->gc) {
    g->gc = gdk_gc_new (g->disp->window);
    gdk_gc_copy (g->gc, g->disp->style->white_gc);
  }
  
  /* Allocate colors. */
  if (!g->colors_allocated) {
    GdkColormap *colormap;
    
    colormap = gdk_window_get_colormap (g->disp->window);
    for (i=0;i<3;i++) 
      gdk_color_alloc (colormap, &(g->colors [i]));    
    
    g->colors_allocated = 1;
  }
  /* Erase Rectangle */
  gdk_gc_set_foreground (g->gc, &(g->colors [0]));
  gdk_draw_rectangle (g->pixmap,
		      g->gc,
		      TRUE, 0, 0,
		      g->disp->allocation.width,
		      g->disp->allocation.height);
  
  /* draw frame */
  gdk_gc_set_foreground (g->gc, &(g->colors [1]));
  gdk_draw_rectangle (g->pixmap,
		      g->gc,
		      FALSE, 0, 0,
		      g->draw_width,
		      g->disp->allocation.height);
  
  dely = g->draw_height / 5;
  for (i = 1; i <5; i++) {
    gint y1 = g->draw_height + 1 - i * dely;
    gdk_draw_line (g->pixmap, g->gc,
		   0, y1, g->draw_width, y1);
  }
  
  for (i = 0; i < g->num_points; i++)
    g->pos [i] = g->draw_height;
 
  gdk_gc_set_line_attributes (g->gc, 2, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_MITER );
  delx = (float)g->draw_width / ( g->num_points - 1);
  gdk_gc_set_foreground (g->gc, &(g->colors [0 + 2]));
  
  for (i = 0; i < g->num_points - 1; i++) {    
    gint x1 = i * delx;
    gint x2 = (i + 1) * delx;
    gint y1 = g->data[i][0] * g->draw_height - 1;
    gint y2 = g->data[i+1][0] * g->draw_height - 1;
    
    if ((g->data[i][0] != -1) && (g->data[i+1][0] != -1))
      gdk_draw_line (g->pixmap, g->gc,
		     g->draw_width - x2, g->pos[i + 1] - y2,
		     g->draw_width - x1, g->pos[i] - y1);
    g->pos [i] -= g->data [i][0];
  }
  g->pos[g->num_points - 1] -= g->data [g->num_points - 1][0];
  
  gdk_gc_set_line_attributes (g->gc, 1, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_MITER );
  
  gdk_draw_pixmap (g->disp->window,
		   g->disp->style->fg_gc [GTK_WIDGET_STATE(g->disp)],
		   g->pixmap,
		   0, 0,
		   0, 0,
		   g->disp->allocation.width,
		   g->disp->allocation.height);
}

static void get_memory(gfloat * data,
		       LoadGraph *g) {
  STATS_CS_MESSAGE * statMsg;
  CS_HEADER csHdr;
  unsigned int count;
  unsigned int i;
  int mpos;
  long long value = -1;
  
  data[0] = -1; /* invalid */
  csHdr.size 
    = htons(sizeof(CS_HEADER));
  csHdr.tcpType
    = htons(STATS_CS_PROTO_GET_STATISTICS);
  if (SYSERR == writeToSocket(g->sock,
			      &csHdr)) {
    fprintf(stderr,
	    "Error sending request for statistics to gnunetd.\n");
    return;
  }
  statMsg 
    = MALLOC(MAX_BUFFER_SIZE);
  statMsg->totalCounters 
    = htonl(1); /* to ensure we enter the loop */
  count = 0;
  while ( count < ntohl(statMsg->totalCounters) ) {
    if (SYSERR == readFromSocket(g->sock,
				 (CS_HEADER**)&statMsg)) {
      fprintf(stderr,
	      "Error receiving reply for statistics from gnunetd.\n");
      FREE(statMsg);
      return;    
    }
    if (ntohs(statMsg->header.size) < sizeof(STATS_CS_MESSAGE)) {
      LOG(LOG_WARNING,
	  "WARNING: received malformed stats message (%d < %d)\n",
	  ntohs(statMsg->header.size), 
	  sizeof(STATS_CS_MESSAGE) );
      break;
    }
    mpos = sizeof(unsigned long long) * ntohl(statMsg->statCounters);
    for (i=0;i<ntohl(statMsg->statCounters);i++) {
      if (mpos+strlen(&((char*)(((STATS_CS_MESSAGE_GENERIC*)statMsg)->values))[mpos])+1 > 
	  ntohs(statMsg->header.size) - sizeof(STATS_CS_MESSAGE)) {
	LOG(LOG_WARNING,
	    "WARNING: received malformed stats message (%d > %d)\n",
	    mpos+strlen(&((char*)(((STATS_CS_MESSAGE_GENERIC*)statMsg)->values))[mpos])+1,
	    ntohs(statMsg->header.size) - sizeof(STATS_CS_MESSAGE));
	break; /* out of bounds! */      
      }
      if (0 == strcmp(&((char*)(((STATS_CS_MESSAGE_GENERIC*)statMsg)->values))[mpos],
		      stats[g->statIdx].statName)) 
	value = ntohll(((STATS_CS_MESSAGE_GENERIC*)statMsg)->values[i]);      
      mpos += strlen(&((char*)(((STATS_CS_MESSAGE_GENERIC*)statMsg)->values))[mpos])+1;
    }    
    count += ntohl(statMsg->statCounters);
  } /* end while */
  FREE(statMsg);

  data[0] = value / 100.0;  
}

/* Updates the load graph when the timeout expires */
static int load_graph_update(LoadGraph *g) {
  guint i;  

  for (i=0;i<g->num_points;i++)
    memcpy(g->odata[i], 
	   g->data[i],
	   g->data_size); 
  get_memory(g->data[0], g);
  for (i=0;i<g->num_points-1;i++)
    g->data[i+1][0] = g->odata[i][0];  
  if (g->draw)
    load_graph_draw (g);  
  return TRUE;
}

static void load_graph_unalloc (LoadGraph *g) {
  int i;
  if (!g->allocated)
    return;
  for (i = 0; i < g->num_points; i++) {
    g_free (g->data [i]);
    g_free (g->odata [i]);
  }
  g_free (g->data);
  g_free (g->odata);
  g_free (g->pos);
  g->pos = NULL;
  g->data = g->odata = NULL;
  if (g->pixmap) {
    gdk_pixmap_unref (g->pixmap);
    g->pixmap = NULL;
  }
  g->allocated = FALSE;
}

static void load_graph_alloc (LoadGraph *g) {
  int i;

  if (g->allocated)
    return;
  
  g->data = g_new0 (gfloat *, g->num_points);
  g->odata = g_new0 (gfloat *, g->num_points);
  g->pos = g_new0 (guint, g->num_points);
  g->data_size = sizeof (gfloat);  
  for (i = 0; i < g->num_points; i++) {
    g->data[i] = g_malloc0(g->data_size);
    g->odata[i] = g_malloc0(g->data_size);
  }  
  for (i=0;i<g->num_points;i++) 
    g->data[i][0] = -1;
  g->allocated = TRUE;
}

static gint load_graph_configure(GtkWidget *widget, 
				 GdkEventConfigure *event,
				 gpointer data_ptr) {
  LoadGraph *c = (LoadGraph *) data_ptr;
  
  if (c->pixmap) {
    gdk_pixmap_unref (c->pixmap);
    c->pixmap = NULL;
  }
  
  if (!c->pixmap)
    c->pixmap = gdk_pixmap_new (widget->window,
				widget->allocation.width,
				widget->allocation.height,
				gtk_widget_get_visual (c->disp)->depth);
  
  gdk_draw_rectangle (c->pixmap,
		      widget->style->black_gc,
		      TRUE, 0,0,
		      widget->allocation.width,
		      widget->allocation.height);
  gdk_draw_pixmap (widget->window,
		   c->disp->style->fg_gc [GTK_WIDGET_STATE(widget)],
		   c->pixmap,
		   0, 0,
		   0, 0,
		   c->disp->allocation.width,
		   c->disp->allocation.height);
  
  load_graph_draw (c);
  
  return TRUE;
  event = NULL;
}

static gint load_graph_expose (GtkWidget *widget,
			       GdkEventExpose *event,
			       gpointer data_ptr) {
  LoadGraph *g = (LoadGraph *) data_ptr;
  
  gdk_draw_pixmap (widget->window,
		   widget->style->fg_gc [GTK_WIDGET_STATE(widget)],
		   g->pixmap,
		   event->area.x, event->area.y,
		   event->area.x, event->area.y,
		   event->area.width, event->area.height);
  return FALSE;
}

static void load_graph_stop (LoadGraph *g) {
  if (g->timer_index != -1) {
    gtk_timeout_remove (g->timer_index);
    g->timer_index = -1;
  }
  if (!g)
    return;
  g->draw = FALSE;
}

static void load_graph_destroy(GtkWidget *widget, 
			       gpointer data_ptr) {
  LoadGraph *g = (LoadGraph *) data_ptr;  
  load_graph_stop(g);  
  if (g->timer_index != -1)
    gtk_timeout_remove (g->timer_index);
  if (g->sock != NULL)
    releaseClientSocket(g->sock);
  load_graph_unalloc(g);
}

static LoadGraph * load_graph_new(ProcData *procdata) {
  LoadGraph *g;

  g = g_new0(LoadGraph, 1);
  g->sock = getClientSocket();
  g->statIdx = procdata->statIdx;
  g->speed = procdata->config.graph_update_interval;
  g->num_points = 200;
  g->colors = g_new0 (GdkColor, 3);
  g->colors[0] = procdata->config.bg_color;
  g->colors[1] = procdata->config.frame_color;
  g->colors[2] = procdata->config.mem_color;
  g->timer_index = -1;
  g->draw = FALSE;
  g->main_widget = gtk_vbox_new (FALSE, FALSE);
  gtk_widget_show (g->main_widget);
  g->disp = gtk_drawing_area_new();
  gtk_widget_show (g->disp);
  gtk_signal_connect (GTK_OBJECT (g->disp),
		      "expose_event",
		      GTK_SIGNAL_FUNC (load_graph_expose), g);
  gtk_signal_connect (GTK_OBJECT(g->disp), 
		      "configure_event",
		      GTK_SIGNAL_FUNC (load_graph_configure), g);
  gtk_signal_connect (GTK_OBJECT(g->disp),
		      "destroy",
		      GTK_SIGNAL_FUNC (load_graph_destroy), g); 
  gtk_widget_set_events (g->disp, GDK_EXPOSURE_MASK);  
  gtk_box_pack_start (GTK_BOX (g->main_widget), g->disp, TRUE, TRUE, 0);  
  load_graph_alloc(g);  
  gtk_widget_show_all (g->main_widget);  
  g->timer_index = gtk_timeout_add(g->speed,
				   (GtkFunction) load_graph_update, g);
  
  return g;
}

static void load_graph_start(LoadGraph *g) {
  if (!g)
    return;
  
  if (g->timer_index == -1)
    g->timer_index = gtk_timeout_add(g->speed,
				     (GtkFunction) load_graph_update, g);
  
  g->draw = TRUE;
}

static GtkWidget * create_sys_view(ProcData * procdata) {
  GtkWidget * vbox;
  GtkWidget * mem_frame;
  GtkWidget * menu;
  GtkItemFactory * popupFactory;
  LoadGraph * mem_graph;
  
  vbox = gtk_vbox_new (FALSE, 0);

  mem_frame = gtk_frame_new (stats[procdata->statIdx].frameName);
  gtk_container_set_border_width (GTK_CONTAINER (mem_frame), GNOME_PAD_SMALL);
  gtk_box_pack_start (GTK_BOX (vbox), mem_frame, TRUE, TRUE, 0);
  
  mem_graph = load_graph_new(procdata);
  gtk_container_add (GTK_CONTAINER (mem_frame), mem_graph->main_widget);
  gtk_container_set_border_width (GTK_CONTAINER (mem_graph->main_widget),
				  GNOME_PAD_SMALL);
  procdata->mem_graph = mem_graph;

  /* FIXME: this should add a right-button popup-menu 
     but does not. Somehow popupCallback() never gets called,
     i.e. event does not occur or has been blocked or 
     overridden somewhere. Trying to connect to some other
     widget instead of vbox stays equally unresponsive */
  popupFactory = gtk_item_factory_new (GTK_TYPE_MENU, "<main>",
  				       NULL);
  gtk_item_factory_create_items(popupFactory,
  				statWindowMenuItems,
				statWindowMenu,
				NULL);
  menu = gtk_item_factory_get_widget(popupFactory, "<main>");
  gtk_signal_connect(GTK_OBJECT(vbox),
 		     "event",
		     GTK_SIGNAL_FUNC(popupCallback),
		     menu);
  gtk_widget_show_all (vbox);
  
  return vbox;
}


static GtkWidget * create_main_window(int stat) {
  GtkWidget *sys_box;
  ProcData procdata;

  memset(&procdata, 0, sizeof(ProcData));
  procdata.config.graph_update_interval = 1000; /* every second */
  procdata.statIdx = stat;
  gdk_color_parse("red",
		  &procdata.config.mem_color);
  gdk_color_parse("black",
		  &procdata.config.bg_color);
  gdk_color_parse("blue",
		  &procdata.config.frame_color);
  sys_box = create_sys_view(&procdata);
  load_graph_start (procdata.mem_graph); 
  return sys_box;
}




/**
 * Display the statistics.
 */
void displayStatistics(GtkWidget * widget,
		       gpointer data) {
  int dptr = (int) data;
  addToNotebook(stats[dptr].paneName,
		create_main_window(dptr));
}
 
/* end of statistics.c */
